import inspect
from abc import ABC, abstractmethod
from typing import List

from workspaces.capability.enums import CapabilityStepType
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityIF,
    CapabilityRequestIF, CapabilityVersionIF,
)
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import AbstractFile

CapabilityName = str


class CapabilityInfoIF(ABC):
    """
    Data access object that can look up and record information about capabilities
    and capability requests.
    """

    @abstractmethod
    def create_capability(
        self, name: CapabilityName, steps: List[CapabilityStepType], max_jobs: int
    ) -> CapabilityIF:
        """
        Create new capability and save it in the database
        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Integer identifier for the capability
        """
        raise NotImplementedError

    @abstractmethod
    def create_capability_request(
        self,
        capability_name: str,
        parameters: List[ParameterIF] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequestIF:
        """
        Create new capability request and save it in the database
        :param capability_name: name of the requested capability
        :param parameters: List
        :param future_products:
        :return: Integer identifier of the request
        """
        raise NotImplementedError

    @abstractmethod
    def create_capability_version(
            self,
            capability_request_id: int,
            parameters: List[ParameterIF] = None,
    ) -> CapabilityVersionIF:
        """
        Create new capability request and save it in the database
        :param capability_request_id: id of the capability request to attach new version to
        :param parameters: List
        :return: Integer identifier of the version
        """
        raise NotImplementedError

    @abstractmethod
    def create_execution(self, request: CapabilityRequestIF) -> CapabilityExecutionIF:
        """
        Create new execution record for a request and save it in the database
        :param request: ID of the capability request
        :return: Integer identifier for the record
        """
        raise NotImplementedError

    @abstractmethod
    def lookup_capability(self, capability_name: str) -> CapabilityIF:
        raise NotImplementedError

    @abstractmethod
    def lookup_capability_request(self, request_id) -> CapabilityRequestIF:
        raise NotImplementedError

    @abstractmethod
    def lookup_execution(self, execution_id: int) -> CapabilityExecutionIF:
        raise NotImplementedError

    @abstractmethod
    def lookup_execution_by_workflow_request_id(
        self, workflow_request_id: int
    ) -> CapabilityExecutionIF:
        raise NotImplementedError

    @abstractmethod
    def save_execution(self, execution: CapabilityExecutionIF):
        pass

    @abstractmethod
    def save_entity(self, entity):
        pass


class CapabilityQueueIF(ABC):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    @abstractmethod
    def enqueue(self, execution: CapabilityExecutionIF, priority: int):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def dequeue(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def pause(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")


class CapabilityEngineIF(ABC):
    """
    Executes a prepare and run workflow step of a capability
    """

    @abstractmethod
    def load_engine(self, execution: CapabilityExecutionIF):
        pass

    @abstractmethod
    def execute(self):
        pass

    @abstractmethod
    def submit_workflow_request(
        self, execution_id: int, workflow_name: str, workflow_args: dict, files: List[AbstractFile]
    ):
        pass


class CapabilityServiceIF(ABC):
    """
    The capability service: clients access this to request capability runs
    """

    pass
